﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.DataAnalysis;
using HeuristicLab.Data;

namespace HeuristicLab.Functions {
  class BakedFunctionTree : ItemBase, IFunctionTree {
    private List<int> code;
    private List<double> data;
    private const int ADDITION = 10010;
    private const int AND = 10020;
    private const int AVERAGE = 10030;
    private const int CONSTANT = 10040;
    private const int COSINUS = 10050;
    private const int DIVISION = 10060;
    private const int EQU = 10070;
    private const int EXP = 10080;
    private const int GT = 10090;
    private const int IFTE = 10100;
    private const int LT = 10110;
    private const int LOG = 10120;
    private const int MULTIPLICATION = 10130;
    private const int NOT = 10140;
    private const int OR = 10150;
    private const int POWER = 10160;
    private const int SIGNUM = 10170;
    private const int SINUS = 10180;
    private const int SQRT = 10190;
    private const int SUBSTRACTION = 10200;
    private const int TANGENS = 10210;
    private const int VARIABLE = 10220;
    private const int XOR = 10230;

    private static int nextFunctionSymbol = 10240;
    private static Dictionary<int, IFunction> symbolTable;
    private static Dictionary<IFunction, int> reverseSymbolTable;
    private static Dictionary<Type, int> staticTypes;

    static BakedFunctionTree() {
      symbolTable = new Dictionary<int, IFunction>();
      reverseSymbolTable = new Dictionary<IFunction, int>();
      staticTypes = new Dictionary<Type, int>();
      staticTypes[typeof(Addition)] = ADDITION;
      staticTypes[typeof(And)] = AND;
      staticTypes[typeof(Average)] = AVERAGE;
      staticTypes[typeof(Constant)] = CONSTANT;
      staticTypes[typeof(Cosinus)] = COSINUS;
      staticTypes[typeof(Division)] = DIVISION;
      staticTypes[typeof(Equal)] = EQU;
      staticTypes[typeof(Exponential)] = EXP;
      staticTypes[typeof(GreaterThan)] = GT;
      staticTypes[typeof(IfThenElse)] = IFTE;
      staticTypes[typeof(LessThan)] = LT;
      staticTypes[typeof(Logarithm)] = LOG;
      staticTypes[typeof(Multiplication)] = MULTIPLICATION;
      staticTypes[typeof(Not)] = NOT;
      staticTypes[typeof(Or)] = OR;
      staticTypes[typeof(Power)] = POWER;
      staticTypes[typeof(Signum)] = SIGNUM;
      staticTypes[typeof(Sinus)] = SINUS;
      staticTypes[typeof(Sqrt)] = SQRT;
      staticTypes[typeof(Substraction)] = SUBSTRACTION;
      staticTypes[typeof(Tangens)] = TANGENS;
      staticTypes[typeof(Variable)] = VARIABLE;
      staticTypes[typeof(Xor)] = XOR;
    }

    internal BakedFunctionTree() {
      code = new List<int>();
      data = new List<double>();
    }

    internal BakedFunctionTree(IFunction function)
      : this() {
      code.Add(0);
      code.Add(MapFunction(function));
      code.Add(0);
      treesExpanded = true;
      subTrees = new List<IFunctionTree>();
      variables = new List<IVariable>();
      variablesExpanded = true;
      foreach(IVariableInfo variableInfo in function.VariableInfos) {
        if(variableInfo.Local) {
          variables.Add((IVariable)function.GetVariable(variableInfo.FormalName).Clone());
        }
      }
    }

    internal BakedFunctionTree(IFunctionTree tree)
      : this() {
      code.Add(0);
      code.Add(MapFunction(tree.Function));
      code.Add(tree.LocalVariables.Count);
      foreach(IVariable variable in tree.LocalVariables) {
        IItem value = variable.Value;
        data.Add(GetDoubleValue(value));
      }
      foreach(IFunctionTree subTree in tree.SubTrees) {
        AddSubTree(new BakedFunctionTree(subTree));
      }
    }

    private double GetDoubleValue(IItem value) {
      if(value is DoubleData) {
        return ((DoubleData)value).Data;
      } else if(value is ConstrainedDoubleData) {
        return ((ConstrainedDoubleData)value).Data;
      } else if(value is IntData) {
        return ((IntData)value).Data;
      } else if(value is ConstrainedIntData) {
        return ((ConstrainedIntData)value).Data;
      } else throw new NotSupportedException("Invalid datatype of local variable for GP");
    }

    private int MapFunction(IFunction function) {
      if(!reverseSymbolTable.ContainsKey(function)) {
        int curFunctionSymbol;
        if(staticTypes.ContainsKey(function.GetType())) curFunctionSymbol = staticTypes[function.GetType()];
        else {
          curFunctionSymbol = nextFunctionSymbol;
          nextFunctionSymbol++;
        }
        reverseSymbolTable[function] = curFunctionSymbol;
        symbolTable[curFunctionSymbol] = function;
      }
      return reverseSymbolTable[function];
    }

    private void BranchLength(int branchRoot, out int codeLength, out int dataLength) {
      int arity = code[branchRoot];
      int nLocalVariables = code[branchRoot + 2];
      codeLength = 3;
      dataLength = nLocalVariables;
      int subBranchStart = branchRoot + codeLength;
      for(int i = 0; i < arity; i++) {
        int branchCodeLength;
        int branchDataLength;
        BranchLength(subBranchStart, out branchCodeLength, out branchDataLength);
        subBranchStart += branchCodeLength;
        codeLength += branchCodeLength;
        dataLength += branchDataLength;
      }
    }

    private void FlattenTrees() {
      if(treesExpanded) {
        code[0] = subTrees.Count;
        foreach(BakedFunctionTree subTree in subTrees) {
          subTree.FlattenVariables();
          subTree.FlattenTrees();
          code.AddRange(subTree.code);
          data.AddRange(subTree.data);
        }
        treesExpanded = false;
        subTrees = null;
      }
    }

    private void FlattenVariables() {
      if(variablesExpanded) {
        code[2] = variables.Count;
        foreach(IVariable variable in variables) {
          data.Add(GetDoubleValue(variable.Value));
        }
        variablesExpanded = false;
        variables = null;
      }
    }

    private bool treesExpanded = false;
    private List<IFunctionTree> subTrees;
    public IList<IFunctionTree> SubTrees {
      get {
        if(!treesExpanded) {
          subTrees = new List<IFunctionTree>();
          int arity = code[0];
          int nLocalVariables = code[2];
          int branchIndex = 3;
          int dataIndex = nLocalVariables; // skip my local variables to reach the local variables of the first branch
          for(int i = 0; i < arity; i++) {
            BakedFunctionTree subTree = new BakedFunctionTree();
            int codeLength;
            int dataLength;
            BranchLength(branchIndex, out codeLength, out dataLength);
            subTree.code = code.GetRange(branchIndex, codeLength);
            subTree.data = data.GetRange(dataIndex, dataLength);
            branchIndex += codeLength;
            dataIndex += dataLength;
            subTrees.Add(subTree);
          }
          treesExpanded = true;
          code.RemoveRange(3, code.Count - 3);
          code[0] = 0;
          data.RemoveRange(nLocalVariables, data.Count - nLocalVariables);
        }
        return subTrees;
      }
    }

    private bool variablesExpanded = false;
    private List<IVariable> variables;
    public ICollection<IVariable> LocalVariables {
      get {
        if(!variablesExpanded) {
          variables = new List<IVariable>();
          IFunction function = symbolTable[code[1]];
          int localVariableIndex = 0;
          foreach(IVariableInfo variableInfo in function.VariableInfos) {
            if(variableInfo.Local) {
              IVariable clone = (IVariable)function.GetVariable(variableInfo.FormalName).Clone();
              IItem value = clone.Value;
              if(value is ConstrainedDoubleData) {
                ((ConstrainedDoubleData)value).Data = data[localVariableIndex];
              } else if(value is ConstrainedIntData) {
                ((ConstrainedIntData)value).Data = (int)data[localVariableIndex];
              } else if(value is DoubleData) {
                ((DoubleData)value).Data = data[localVariableIndex];
              } else if(value is IntData) {
                ((IntData)value).Data = (int)data[localVariableIndex];
              } else throw new NotSupportedException("Invalid local variable type for GP.");
              variables.Add(clone);
              localVariableIndex++;
            }
          }
          variablesExpanded = true;
          code[2] = 0;
          data.RemoveRange(0, variables.Count);
        }
        return variables;
      }
    }

    public IFunction Function {
      get { return symbolTable[code[1]]; }
    }

    public IVariable GetLocalVariable(string name) {
      foreach(IVariable var in LocalVariables) {
        if(var.Name == name) return var;
      }
      return null;
    }

    public void AddVariable(IVariable variable) {
      throw new NotSupportedException();
    }

    public void RemoveVariable(string name) {
      throw new NotSupportedException();
    }

    public void AddSubTree(IFunctionTree tree) {
      if(!treesExpanded) throw new InvalidOperationException();
      subTrees.Add(tree);
    }

    public void InsertSubTree(int index, IFunctionTree tree) {
      if(!treesExpanded) throw new InvalidOperationException();
      subTrees.Insert(index, tree);
    }

    public void RemoveSubTree(int index) {
      // sanity check
      if(!treesExpanded) throw new InvalidOperationException();
      subTrees.RemoveAt(index);
    }

    private int PC;
    private int DP;
    private int[] codeArr;
    private double[] dataArr;
    private Dataset dataset;
    private int sampleIndex;
    public double Evaluate(Dataset dataset, int sampleIndex) {
      PC = 0;
      DP = 0;
      FlattenVariables();
      FlattenTrees();
      if(codeArr == null) {
        codeArr = new int[code.Count];
        dataArr = new double[data.Count];
        code.CopyTo(codeArr);
        data.CopyTo(dataArr);
      }
      this.sampleIndex = sampleIndex;
      this.dataset = dataset;
      return EvaluateBakedCode();
    }

    private double EvaluateBakedCode() {
      int arity = codeArr[PC++];
      int functionSymbol = codeArr[PC++];
      int nLocalVariables = codeArr[PC++];
      switch(functionSymbol) {
        case VARIABLE: {
            int var = (int)dataArr[DP++];
            double weight = dataArr[DP++];
            int offset = (int)dataArr[DP++];
            return weight * dataset.GetValue(sampleIndex + offset, var);
          }
        case CONSTANT: {
            double value = dataArr[DP++];
            return value;
          }
        case MULTIPLICATION: {
            double result = 1.0;
            for(int i = 0; i < arity; i++) {
              result *= EvaluateBakedCode();
            }
            return result;
          }
        case ADDITION: {
            double sum = 0.0;
            for(int i = 0; i < arity; i++) {
              sum += EvaluateBakedCode();
            }
            return sum;
          }
        case SUBSTRACTION: {
            if(arity == 1) {
              return -EvaluateBakedCode();
            } else {
              double result = EvaluateBakedCode();
              for(int i = 1; i < arity; i++) {
                result -= EvaluateBakedCode();
              }
              return result;
            }
          }
        case DIVISION: {
            if(arity == 1) {
              double divisor = EvaluateBakedCode();
              if(divisor == 0) return 0;
              else return 1.0 / divisor;
            } else {
              double result = EvaluateBakedCode();
              for(int i = 1; i < arity; i++) {
                double divisor = EvaluateBakedCode();
                if(divisor == 0) result = 0;
                else result /= divisor;
              }
              return result;
            }
          }
        case AVERAGE: {
            double sum = 0.0;
            for(int i = 0; i < arity; i++) {
              sum += EvaluateBakedCode();
            }
            return sum / arity;
          }
        case COSINUS: {
            return Math.Cos(EvaluateBakedCode());
          }
        case SINUS: {
            return Math.Sin(EvaluateBakedCode());
          }
        case EXP: {
            return Math.Exp(EvaluateBakedCode());
          }
        case LOG: {
            return Math.Log(EvaluateBakedCode());
          }
        case POWER: {
            double x = EvaluateBakedCode();
            double p = EvaluateBakedCode();
            return Math.Pow(x, p);
          }
        case SIGNUM: {
            // protected signum
            double value = EvaluateBakedCode();
            if(value < 0) return -1;
            if(value > 0) return 1;
            return 0;
          }
        case SQRT: {
            return Math.Sqrt(EvaluateBakedCode());
          }
        case TANGENS: {
            return Math.Tan(EvaluateBakedCode());
          }
        case AND: {
            double result = 1.0;
            // have to evaluate all sub-trees, skipping would probably not lead to a big gain because 
            // we have to iterate over the linear structure anyway
            for(int i = 0; i < arity; i++) {
              double x = Math.Round(EvaluateBakedCode());
              if(x == 0) result *= 0;
              else if(x == 1.0) result *= 1.0;
              else result *= double.NaN;
            }
            return result;
          }
        case EQU: {
            double x = EvaluateBakedCode();
            double y = EvaluateBakedCode();
            if(x == y) return 1.0; else return 0.0;
          }
        case GT: {
            double x = EvaluateBakedCode();
            double y = EvaluateBakedCode();
            if(x > y) return 1.0;
            else return 0.0;
          }
        case IFTE: {
            double condition = Math.Round(EvaluateBakedCode());
            double x = EvaluateBakedCode();
            double y = EvaluateBakedCode();
            if(condition < .5) return x;
            else if(condition >= .5) return y;
            else return double.NaN;
          }
        case LT: {
            double x = EvaluateBakedCode();
            double y = EvaluateBakedCode();
            if(x < y) return 1.0;
            else return 0.0;
          }
        case NOT: {
            double result = Math.Round(EvaluateBakedCode());
            if(result == 0.0) return 1.0;
            else if(result == 1.0) return 0.0;
            else return double.NaN;
          }
        case OR: {
            double result = 0.0; // default is false
            for(int i = 0; i < arity; i++) {
              double x = Math.Round(EvaluateBakedCode());
              if(x == 1.0 && result == 0.0) result = 1.0; // found first true (1.0) => set to true
              else if(x != 0.0) result = double.NaN; // if it was not true it can only be false (0.0) all other cases are undefined => (NaN)
            }
            return result;
          }
        case XOR: {
            double x = Math.Round(EvaluateBakedCode());
            double y = Math.Round(EvaluateBakedCode());
            if(x == 0.0 && y == 0.0) return 0.0;
            if(x == 1.0 && y == 0.0) return 1.0;
            if(x == 0.0 && y == 1.0) return 1.0;
            if(x == 1.0 && y == 1.0) return 0.0;
            return double.NaN;
          }
        default: {
            IFunction function = symbolTable[functionSymbol];
            double[] args = new double[nLocalVariables + arity];
            for(int i = 0; i < nLocalVariables; i++) {
              args[i] = dataArr[DP++];
            }
            for(int j = 0; j < arity; j++) {
              args[nLocalVariables + j] = EvaluateBakedCode();
            }
            return function.Apply(dataset, sampleIndex, args);
          }
      }
    }

    public override System.Xml.XmlNode GetXmlNode(string name, System.Xml.XmlDocument document, IDictionary<Guid, IStorable> persistedObjects) {
      throw new NotImplementedException();
    }

    public override void Populate(System.Xml.XmlNode node, IDictionary<Guid, IStorable> restoredObjects) {
      throw new NotImplementedException();
    }

    public override object Clone(IDictionary<Guid, object> clonedObjects) {
      BakedFunctionTree clone = new BakedFunctionTree();
      if(treesExpanded || variablesExpanded) throw new InvalidOperationException(); // sanity check
      clone.code.AddRange(code);
      clone.data.AddRange(data);
      return clone;
    }

    public override IView CreateView() {
      return new FunctionTreeView(this);
    }
  }
}
